﻿/*
	Actor.cs
	Copyright 2016, OSTRA / Takeshi Yoneki
*/
using System;
using System.Collections.Generic;
using System.Threading;

namespace PictureSetViewLite {
    // 簡易Actor
    public class Actor {
        Queue<Object> cue = new Queue<Object>();
        Object cueLock = new Object();
        Object res = null;
        Object resLock = new Object();
		Object sendLock = new Object();
		Thread th = null;
        bool working = false;
        bool stopRequest = false;

        public Actor() {
        }

        public Actor(ParameterizedThreadStart fn) {
            Create(fn);
        }

        // 開始
        public Actor Create(ParameterizedThreadStart fn) {
            th = new Thread(fn);
            working = true;
            th.Start(this);
            return this;
        }

        // メッセージの非同期送信
        public void Post(Object msg) {
            lock (cueLock) {
                cue.Enqueue(msg);
                if (!working) {
                    working = true;
                    Monitor.Pulse(cueLock);
                }
            }
            Thread.Yield();
        }

        // メッセージの同期送信
        public Object Send(Object msg) {
			// 複数スレッドからのSendの競合がある。
			lock (sendLock) {
				lock (resLock) {
					Post(msg);
					Monitor.Wait(resLock, Timeout.Infinite);
					return res;
				}
			}
		}

		// メッセージの受信
		// 戻り値がnullのときはスレッド終了要求。
		public Object Receive() {
            Thread.Yield();
            lock (cueLock) {
				if (cue.Count == 0) {
					working = false;
					Monitor.Wait(cueLock, Timeout.Infinite);
				}
                stopRequest = false;
                return cue.Dequeue();
			}
        }

        public void Reply(Object result) {
            lock (resLock) {
                res = result;
                Monitor.Pulse(resLock);
            }
        }

        public bool HasCue() {
            lock (cueLock) {
                return cue.Count > 0;
            }
        }

        public bool IsWorking() {
            lock (cueLock) {
                return working;
            }
        }

        public void StopRequest() {
            lock (cueLock) {
                stopRequest = true;
            }
        }

        public bool IsStopRequested() {
            lock (cueLock) {
                return stopRequest;
            }
        }

        // スレッドを抜ける。
        public void Quit() {
            Post(null);
            Thread.Yield();
        }

        public void Join() {
            th.Join();
            th = null;
        }
    }

}