﻿/*
	App.xaml.cs
	Copyright 2016, OSTRA / Takeshi Yoneki
*/
using System;
using System.Windows;
using System.Threading;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Channels;
using System.Runtime.Remoting.Channels.Ipc;

namespace PictureSetViewLite {
	// プロセス間通信でのやり取りに使う型。
    public class RemoteCall: MarshalByRefObject {
        public void Apply(String[] args) {
            // App.Current.MainWindowはプロセス間通信のスレッドでは使えないようだ。
            if (App.mainView != null) {
                // このスレッドはUIスレッドではない。
                App.mainView.Dispatcher.Invoke(() => {
                    if (args.Length > 0) {
                        App.mainView.OpenFileFresh(args[0]);
                    }
                    App.mainView.Act();
                });
            }
        }
    }

    /// <summary>
    /// App.xaml の相互作用ロジック
    /// </summary>
    public partial class App : Application {
        public const String APPL_NAME = "PictureSetViewLite";
        // メインウィンドウのコンストラクタで明示的に値を入れる。
        public static MainWindow mainView = null;
		public static string[] args = new string[0];
		Mutex mutex = new Mutex(false, APPL_NAME);

        private void Application_Startup(object sender, StartupEventArgs e) {
            // Mutexを使って多重起動時の分岐をする。
            //App.Current.MainWindow
            if (mutex.WaitOne(0, false)) {
                // 最初の起動ならプロセス間通信のサーバを立てる。
                IpcServerChannel serverChannel = new IpcServerChannel(APPL_NAME);
                ChannelServices.RegisterChannel(serverChannel, true);
                RemotingConfiguration.RegisterWellKnownServiceType(typeof(RemoteCall), "RemoteCall", WellKnownObjectMode.Singleton);
                args = e.Args;
            }
            else {
                // 二度目以降の起動ならプロセス間通信サーバにコマンドラインパラメタを投げて終了する。
                IpcClientChannel clientChannel = new IpcClientChannel();
                ChannelServices.RegisterChannel(clientChannel, true);
                RemotingConfiguration.RegisterWellKnownClientType(typeof(RemoteCall), "ipc://" + APPL_NAME + "/RemoteCall");
                var remote = new RemoteCall();
                remote.Apply(e.Args);
                Shutdown();
            }
        }
    }
}
