﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Xml;
using System.IO;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.ComponentModel;
using System.Globalization;

namespace PictureSetViewLite {
	public class FilesScriptItem {
		public String Name { get; set; }
		public String command = "";
		public String param = "";
        public String encoding = "";

        static string GetSettingDirPath() {
			return Path.Combine(System.Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "OSTRA");
		}

        public static string GetFilesScriptsPath() {
            string moduleFilePath = Process.GetCurrentProcess().MainModule.FileName;
            string dirPath = Path.GetDirectoryName(moduleFilePath);
            string filesScriptsPath = Path.Combine(dirPath, "FilesScripts");
            return filesScriptsPath;
        }

        static string GetSettingFilePath() {
            return Path.Combine(GetFilesScriptsPath(), "FilesScriptsDef.xml");
            //return Path.Combine(GetSettingDirPath(), "FilesScriptsDef.xml");
		}

		public static List<FilesScriptItem> LoadFilesScriptsDef() {
			var result = new List<FilesScriptItem>();
			var path = GetSettingFilePath();
			if (File.Exists(path)) {
				XmlDocument doc = new XmlDocument();
				doc.Load(path);
				var root = doc.DocumentElement;
				if (root != null) {
					foreach (XmlNode node in root.GetElementsByTagName("item")) {
						var fs = new FilesScriptItem();
						XmlElement item = (XmlElement)node;
						fs.Name = item.GetAttribute("name");
						fs.command = item.GetAttribute("command");
						fs.param = item.GetAttribute("param");
                        fs.encoding = item.GetAttribute("encoding");
                        result.Add(fs);
					}
				}
			}
			return result;
		}
	}

	public class FileItem {
		public String Name { get; set; }
		public String path;
	}

	public class FilesDlgModel: INotifyPropertyChanged {
		public event PropertyChangedEventHandler PropertyChanged;
		public List<FilesScriptItem> FilesScriptList { get; set; }
		List<FileItem> FileItemListInstance = new List<FileItem>();
		public List<FileItem> FileItemList { 
			get { 
				return FileItemListInstance;
			} 
			set {
				FileItemListInstance = value;
				PropertyChanged(this, new PropertyChangedEventArgs("FileItemList"));
			} 
		}
		//
		public FilesDlgModel() {
			FilesScriptList = FilesScriptItem.LoadFilesScriptsDef();
		}
	}

	/// <summary>
	/// FilesDlg.xaml の相互作用ロジック
	/// </summary>
	public partial class FilesDlg : Window {
		public MainWindow mainView;
        WindowState savedState = Prefs.data.filesDlgWindowState;
		public RoutedUICommand uiOpenCommand { get; set; }
		public RoutedUICommand uiCloseCommand { get; set; }
		public FilesDlg() {
			InitializeComponent();
			//
			if (Prefs.data.filesDlgWidth > 0.0 && Prefs.data.filesDlgHeight > 0.0) {
				Top = Prefs.data.filesDlgTop;
				Left = Prefs.data.filesDlgLeft;
				Width = Prefs.data.filesDlgWidth;
				Height = Prefs.data.filesDlgHeight;
                WindowState = Prefs.data.filesDlgWindowState;
            }
			if (cbSelectScript.Items.Count > Prefs.data.filesScriptIndex) {
				cbSelectScript.SelectedIndex = Prefs.data.filesScriptIndex;
			}
            else {
                cbSelectScript.SelectedIndex = 0;
            }
            cbClose.IsChecked = Prefs.data.filesDlgClose;
			//
			uiOpenCommand = CreateMenuCommand("open", new KeyGesture[] { new KeyGesture(Key.Enter) });
			CommandBindings.Add(new CommandBinding(uiOpenCommand, OpenCommand, CanAlwaysTrue));
			uiCloseCommand = CreateMenuCommand("close", new KeyGesture[] { new KeyGesture(Key.Escape) });
			CommandBindings.Add(new CommandBinding(uiCloseCommand, CloseCommand, CanAlwaysTrue));
			//
		}

		private RoutedUICommand CreateMenuCommand(String name, KeyGesture[] shortcuts) {
			var playCmdIG = new InputGestureCollection();
			foreach (KeyGesture shortcut in shortcuts) {
				playCmdIG.Add(shortcut);
			}
			return new RoutedUICommand(name, name, typeof(MainWindow), playCmdIG);
		}

		// リストボックスのSelectedItemを設定すると一見適切に選択されているようにみえるが、
		// キーボードの矢印キーで移動すると先頭からになってしまう。
		// コンストラクタでフォーカスを与えようとしたがItemContainerGenerator.ContainerFromItemがnullを返すためLoadedで行う。
		// これでは必ずしも解決するわけではないようだ。elmがnullのときもある。
		private int afterLoaded = 0;

		private void Window_Loaded(object sender, RoutedEventArgs e) {
			lbFiles.Focus();
			if (lbFiles.SelectedItem != null && afterLoaded == 0) { 
				var elm = (UIElement)lbFiles.ItemContainerGenerator.ContainerFromItem(lbFiles.SelectedItem);
				if (elm != null) {
					elm.Focus();
				}
			}
			afterLoaded = 1;
		}

		private void Window_LayoutUpdated(object sender, EventArgs e) {
			if (afterLoaded == 1 && lbFiles.SelectedItem != null) {
				var elm = (UIElement)lbFiles.ItemContainerGenerator.ContainerFromItem(lbFiles.SelectedItem);
				if (elm != null) {
					elm.Focus();
					afterLoaded = 2;
				}
			}
		}

		String RunScriptCommand(String command, String param, String encoding) {
			try {
                Encoding enc;
                if (encoding.Length != 0) {
                    enc = Encoding.GetEncoding(encoding);
                }
                else {
                    enc = Encoding.GetEncoding(CultureInfo.CurrentCulture.TextInfo.ANSICodePage);
                }
                var psi = new ProcessStartInfo(command, param);
				psi.UseShellExecute = false;
				psi.RedirectStandardOutput = true;
				psi.RedirectStandardError = true;
				psi.CreateNoWindow = true;
                psi.WorkingDirectory = FilesScriptItem.GetFilesScriptsPath();
                psi.StandardOutputEncoding = enc;
                var proc = Process.Start(psi);
				var t = proc.StandardOutput.ReadToEnd();
				var e = proc.StandardError.ReadToEnd();
				proc.WaitForExit();
                proc.Close();
                return t;
			}
			catch {
				return "";
			}
		}

		private void cbSelectScript_SelectionChanged(object sender, SelectionChangedEventArgs e) {
			var model = (FilesDlgModel)DataContext;
			Prefs.data.filesScriptIndex = cbSelectScript.SelectedIndex;
			var fs = model.FilesScriptList[cbSelectScript.SelectedIndex];
			var pathList = RunScriptCommand(fs.command, fs.param, fs.encoding).Split(new Char[] {'\n'});
			var itemList = new List<FileItem>();
			int selIndex = 0;
			int i = 0;
			foreach (var t in pathList) {
				var path = t.Trim();
				if (path.Length > 0) { 
					var item = new FileItem();
					item.Name = Path.GetFileName(path);
					item.path = path;
                    if (String.Compare(Prefs.data.filePath, path, StringComparison.CurrentCultureIgnoreCase) == 0) { 
						selIndex = i;
					}
					itemList.Add(item);
					i++;
				}
			}
			model.FileItemList = itemList;
			if (lbFiles.Items.Count > selIndex) {
				int scrollIndex = selIndex;
				for (int d = 0; d < 2 && lbFiles.Items.Count - 1 > scrollIndex; d++) {
					scrollIndex++;
				}
				//lbFiles.ScrollIntoView(lbFiles.Items[selIndex]);
				lbFiles.ScrollIntoView(lbFiles.Items[scrollIndex]);
				lbFiles.SelectedItem = lbFiles.Items[selIndex];
			}
		}

		public void CanAlwaysTrue(object sender, CanExecuteRoutedEventArgs e) {
			e.CanExecute = true;
		}

		public void OpenCommand(object sender, ExecutedRoutedEventArgs e) {
			var model = (FilesDlgModel)DataContext;
			Prefs.data.fileItemIndex = lbFiles.SelectedIndex;
			mainView.OpenFileFresh(model.FileItemList[lbFiles.SelectedIndex].path);
			Prefs.data.filesDlgClose = cbClose.IsChecked == true;
			if (Prefs.data.filesDlgClose) {
				Close();
			}
			else {
				mainView.Act();
			}
		}

		public void CloseCommand(object sender, ExecutedRoutedEventArgs e) {
			Close();
		}

		private void lbFiles_MouseDoubleClick(object sender, MouseButtonEventArgs e) {
            var elm = (UIElement)lbFiles.ItemContainerGenerator.ContainerFromItem(lbFiles.SelectedItem);
			if (elm != null && elm.InputHitTest(e.GetPosition(elm)) != null) {
				OpenCommand(sender, null);
			}
		}

		private void Window_Closing(object sender, System.ComponentModel.CancelEventArgs e) {
			if (WindowState == WindowState.Normal) {
				Prefs.data.filesDlgTop = Top;
				Prefs.data.filesDlgLeft = Left;
				Prefs.data.filesDlgWidth = Width;
				Prefs.data.filesDlgHeight = Height;
			}
            if (WindowState != WindowState.Minimized) {
                Prefs.data.filesDlgWindowState = WindowState;
            }
        }

        private void Window_Closed(object sender, EventArgs e) {
            mainView.filesDlg = null;
        }

        public void Act() {
            if (WindowState == WindowState.Minimized) {
                WindowState = savedState;
            }
            Activate();
        }

        private void Window_StateChanged(object sender, EventArgs e) {
            if (WindowState != WindowState.Minimized) { 
                savedState = WindowState;
            }
        }

		private void Window_KeyDown(object sender, KeyEventArgs e) {
			if (e.Key == Key.Q) {
				Close();
				e.Handled = true;
			}
			else if (e.Key == Key.L) {
				mainView.Act();
				e.Handled = true;
			}
		}

		private void lbFiles_SelectionChanged(object sender, SelectionChangedEventArgs e) {
			//if (lbFiles.SelectedItem != null) {
			//	var elm = (UIElement)lbFiles.ItemContainerGenerator.ContainerFromItem(lbFiles.SelectedItem);
			//	if (elm != null) {
			//		//elm.Focus();
			//	}
			//}
		}

	}
}
