﻿/*
	MainWindow.xaml.cs
	Copyright 2016, OSTRA / Takeshi Yoneki
*/
using System;
using System.Windows;
using System.Windows.Input;
using System.Diagnostics;
using Microsoft.Win32;
using System.IO;
using System.Windows.Media.Imaging;
using System.ComponentModel;
using System.Windows.Controls;
using System.Windows.Interop;
using System.Windows.Media;

namespace PictureSetViewLite {
    // ページ進行
	public enum PageTurnRequest {
		CURRENT, NEXT, PREV, NEXT_ONE, PREV_ONE
	}
    // 画面を6分割
	public enum DividedArea {
		TOP_LEFT, TOP_CENTER, TOP_RIGHT, BOTTOM_LEFT, BOTTOM_CENTER, BOTTOM_RIGHT
	}
    // 画像をどう表示しているか
	public enum PageType {
		NONE, ONE, TWO, START, END
	}

	/// <summary>
	/// MainWindow.xaml の相互作用ロジック
	/// </summary>
	public partial class MainWindow : Window {
		public BitmapPool pool = null;
		public RoutedUICommand uiNextCommand { get; set; }
		public RoutedUICommand uiPrevCommand { get; set; }
		public RoutedUICommand uiNextOneCommand { get; set; }
		public RoutedUICommand uiPrevOneCommand { get; set; }
        public RoutedUICommand uiFullCommand { get; set; }
        public JumpDlg jumpDlg = null;
        public FilesDlg filesDlg = null;
		public AboutDlg aboutDlg = null;
		public PageType pageType;
		double magnifX = 1.0;
        double magnifY = 1.0;
        // ViewPictureをJumpDlgから別スレッドで遅延呼び出しするので、
        // そのときメインウィンドウをクリックするなど再入が発生しうる。
        // 再入を無視するためのフラグ。
        bool enterDrawPicture = false;

        // コンストラクタ
        public MainWindow() {
			InitializeComponent();
            // プロセス間通信でのやり取りに使う
            App.mainView = this;

            // キーボードショートカットの登録
			uiNextCommand = CreateMenuCommand("next", new KeyGesture[] { new KeyGesture(Key.Enter), new KeyGesture(Key.Right), new KeyGesture(Key.Down), new KeyGesture(Key.Space) });
			uiPrevCommand = CreateMenuCommand("prev", new KeyGesture[] { new KeyGesture(Key.Enter, ModifierKeys.Shift), new KeyGesture(Key.Left), new KeyGesture(Key.Back), new KeyGesture(Key.Up), new KeyGesture(Key.Space, ModifierKeys.Shift) });
			uiNextOneCommand = CreateMenuCommand("prev", new KeyGesture[] { new KeyGesture(Key.Right, ModifierKeys.Shift) });
			uiPrevOneCommand = CreateMenuCommand("prev", new KeyGesture[] { new KeyGesture(Key.Left, ModifierKeys.Shift) });
            uiFullCommand = CreateMenuCommand("full", new KeyGesture[] { new KeyGesture(Key.Enter, ModifierKeys.Alt) });
            CommandBindings.Add(new CommandBinding(uiNextCommand, NextCommand, CanAlwaysTrue));
			CommandBindings.Add(new CommandBinding(uiPrevCommand, PrevCommand, CanAlwaysTrue));
			CommandBindings.Add(new CommandBinding(uiNextOneCommand, NextOneCommand, CanAlwaysTrue));
			CommandBindings.Add(new CommandBinding(uiPrevOneCommand, PrevOneCommand, CanAlwaysTrue));
            CommandBindings.Add(new CommandBinding(uiFullCommand, FullCommand, CanAlwaysTrue));

            if (Prefs.data.history == null) {
                Prefs.data.history = new System.Collections.Specialized.StringCollection();
            }

            // ウィンドウの位置とサイズ
            if (Prefs.data.windowWidth > 0.0 && Prefs.data.windowHeight > 0.0) {
				Top = Prefs.data.windowTop;
				Left = Prefs.data.windowLeft;
				Width = Prefs.data.windowWidth;
				Height = Prefs.data.windowHeight;
			}
			if (Prefs.data.windowState != WindowState.Minimized) {
				WindowState = Prefs.data.windowState;
			}
			if (Prefs.data.full) {
				Full(Prefs.data.full);
			}
        }

        // ウィンドウの初期処理後
        private void mainView_Loaded(object sender, RoutedEventArgs e) {
            Trace.WriteLine("mainView_Loaded");

			// システム設定のDPI値ないし拡大倍率をもうちょとマシな方法で得たい。
			// なにこの裏技みたいなのは。
			// PresentationSourceから辿る
			// 「このターゲットからレンダリング先デバイスに座標を変換するために使用できる行列を取得します」
			// もひどい。
			// LoadedのタイミングじゃないとPresentationSource.FromVisual(this)がnullだってのもあんまり。
			// そもそもはウィンドウのサイズ情報で得られるのが倍率計算後の値のみってのが最大の問題というか馬鹿なところ。
			// ドットバイドットの情報の扱いが必要なのは全くもってレアケースでもなんでもない。
			// 論理座標を倍率計算後の値にするという過激なことをするなら色々とサポートが必要だろうに。
			// WPFもまだ高DPI対応途上ってことなんだろう。少なくともWPFにすればOKなんてことではないことは解った。
			//var presentationSource = PresentationSource.FromVisual(this);
			//if (presentationSource != null) {
			//    magnifX = presentationSource.CompositionTarget.TransformToDevice.M11;
			//    magnifY = presentationSource.CompositionTarget.TransformToDevice.M22;
			//}
			//using (var graphics = System.Drawing.Graphics.FromHwnd(IntPtr.Zero)) {
			//    magnifX = graphics.DpiX / 96.0;
			//    magnifY = graphics.DpiY / 96.0;
			//}
			// WindowInteropHelperを使えばIntPtr.Zeroなんてドキュメントに書かれてない値を使わなくて済む。
			// 裏技っぽいのは変わらないが。ただし、Loadedのタイミングじゃないとwih.Handleはnull。
			WindowInteropHelper wih = new WindowInteropHelper(this);
			if (wih != null) {
				using (var graphics = System.Drawing.Graphics.FromHwnd(wih.Handle)) {
					magnifX = graphics.DpiX / 96.0;
					magnifY = graphics.DpiY / 96.0;
				}
			}

            // コマンドラインパラメタで指定されたファイルを開く
            bool done = false;
            foreach (String arg in App.args) {
                done = OpenFileFresh(arg);
                if (done) {
                    break;
                }
            }
            // 指定ファイルがなければ前回終了時のファイルを開く
            if (!done) {
                OpenFile(Prefs.data.filePath, Prefs.data.page);
            }
        }

        private RoutedUICommand CreateMenuCommand(String name, KeyGesture[] shortcuts) {
			var playCmdIG = new InputGestureCollection();
			foreach (KeyGesture shortcut in shortcuts) {
				playCmdIG.Add(shortcut);
			}
			return new RoutedUICommand(name, name, typeof(MainWindow), playCmdIG);
		}

        // 画像プールは有効
		public bool IsEnablePool() {
            return pool != null && pool.IsFileExist();
		}

        // 画像プールは有効で、画像ファイルもある
        public bool HasContentPool() {
            return IsEnablePool() && pool.GetCount() > 0;
        }

        // 何か処理中であることをタイトルに表示。
        // ウィンドウに配置されたコントロールでは即表示とはならず、役に立たない。
        void ProgressOnTitle() {
            // UIスレッド以外から呼ばれても表示できるように。
            Dispatcher.Invoke(() => {
                Title = Title + " .";
            });
        }

        void FixTitle(int page) {
            String name = Path.GetFileNameWithoutExtension(Prefs.data.filePath);
            if (HasContentPool()) {
                // 画像番号は0オリジン
                // 表示ページ番号は+1した値
                // Prefs.data.pageは画像番号であって表示ページ番号ではない。
                if (Prefs.data.pageNumber) {
                    Title = String.Format("{0} ({1} / {2})", name, page + 1, pool.GetCount());
                }
                else {
                    Title = name;
                }
            }
            else if (IsEnablePool()) {
                Title = name;
            }
            else {
                Title = App.APPL_NAME;
            }
        }

        void FixPageNumber(int page) {
            if (HasContentPool() && Prefs.data.pageNumber) {
                // 画像番号は0オリジン
                // 表示ページ番号は+1した値
                // Prefs.data.pageは画像番号であって表示ページ番号ではない。
                textblockPage.Text = String.Format("{0}   /   {1}", page + 1, pool.GetCount());
            }
            else {
                textblockPage.Text = "";
            }
        }

        // メインウィンドウへの描画
        // 再入禁止
        public void DrawPicture(PageTurnRequest turn, bool bUiThread = true) {
            if (enterDrawPicture) {
                return;
            }
            enterDrawPicture = true;
            ProgressOnTitle();
			// GetPageBitmapStreamが動いている間にPrefs.data.pageに変更があると怪しいことになる
			// GetPageBitmapStreamにてPrefs.data.pageなどを直接見るのはやめる
			var w = (int)(gridMain.RenderSize.Width * magnifX);
			var h = (int)(gridMain.RenderSize.Height * magnifX);
            var pageSaved = Prefs.data.page;
            var pb = new PageBitmap(pageSaved, pageType);
            var bms = pb.GetPageBitmapStream(pool, turn, w, h, Prefs.data.ltor, Prefs.data.single);
            if (jumpDlg != null && bUiThread) {
                jumpDlg.ChangePageNumber(pb.page);
            }
            // UIスレッド以外から呼ばれても描画できるように。
            Dispatcher.Invoke(() => {
                // nullが得られたとき何か有効なものが表示されているなら描画しない（最後のページだとか）。
                // そうでなければ（何もないという表現も含め）描画する。
                FixTitle(pb.page);
				if (bms == null && HasContentPool() && pb.page == pageSaved) {
				}
				else {
                    BitmapSource bmx = null;
                    if (bms != null) {
                        BitmapImage bmi = new BitmapImage();
                        bmi.BeginInit();
                        bmi.StreamSource = bms;
                        bmi.EndInit();
                        if (Prefs.data.grayScale) {
							bmx = new FormatConvertedBitmap(bmi, PixelFormats.Gray32Float, null, 0.0);
                        }
                        else {
                            bmx = bmi;
                        }
                    }
                    //imageMain.Stretch = Stretch.None;
                    imageMain.Source = bmx;
					FixPageNumber(pb.page);
                }
            });
			if (pool != null) {
				// 画像キャッシュの充填は描画後に行なう。
				pool.Fill(pb.page);
			}
			Prefs.data.page = pb.page;
			pageType = pb.pageType;
			enterDrawPicture = false;
        }

        // 指定ディレクトリまたはファイルの指定ページを開く
        private bool OpenFile(String path, int page) {
            ProgressOnTitle();
            var result = false;
            FixHistory();
			if (pool != null) {
                pool.Join();
            }
			pool = new BitmapPool(path, Prefs.data.withSubdirectory, Prefs.data.alphabetical);
			Prefs.data.page = 0;
			if (IsEnablePool()) {
                Prefs.data.filePath = pool.filePath;
                if (page < pool.GetCount()) {
                    Prefs.data.page = page;
                }
                //else {
                //    Prefs.data.page = 0;
                //}
                if (pool.containerType == BitmapPool.ContainerType.DIR && BitmapPool.HasImageExt(path)) {
                    Prefs.data.page = pool.GetIndexOfFiles(path);
                }
                FixHistory();
                result = true;
				if (jumpDlg != null) {
					jumpDlg.Setup();
				}
			}
			else {
				if (jumpDlg != null) {
					jumpDlg.Close();
				}
			}
			DrawPicture(PageTurnRequest.CURRENT);
            return result;
        }

        // 新規に指定ディレクトリまたはファイル
        public bool OpenFileFresh(String path) {
            pageType = PageType.NONE;
            return OpenFile(path, GetHistoryPage(path));
        }

        // 履歴の調整
        // 履歴は パス|画像番号 という形式の文字列で保存する
        void FixHistory() {
            if (Prefs.data.filePath.Length == 0) {
                return;
            }
			if (!Prefs.data.enableHistory) {
				return;
			}
			String t = Prefs.data.filePath + "|" + Prefs.data.page.ToString();
            int hit = -1;
            var h = Prefs.data.history;
            for (int i = 0; i < h.Count; i++) {
                var ss = h[i].Split(new Char[] { '|' });
                if (String.Compare(ss[0], Prefs.data.filePath, true) == 0) {
                    hit = i;
                    break;
                }
            }
            if (hit >= 0) {
                h.RemoveAt(hit);
            }
            h.Insert(0, t);
            while (h.Count > 16) {
                h.RemoveAt(h.Count - 1);
            }
        }

        // パスの一致した履歴の画像番号を返す
        int GetHistoryPage(String path) {
            var h = Prefs.data.history;
            for (int i = 0; i < h.Count; i++) {
                var ss = h[i].Split(new Char[] { '|' });
                if (String.Compare(ss[0], path, true) == 0) {
                    try {
                        return Int32.Parse(ss[1]);
                    }
                    catch (Exception) {
                        return 0;
                    }
                }
            }
            return 0;
        }

        // 6分割されたどのエリアがクリックされたか
        private DividedArea GetMousePosDividedArea(double w, double h, Point mp) {
			var top = mp.Y < (h / 2);
			if (mp.X < (w / 3))
				return top ? DividedArea.TOP_LEFT : DividedArea.BOTTOM_LEFT;
			else if (mp.X < (w * 2 / 3))
				return top ? DividedArea.TOP_CENTER : DividedArea.BOTTOM_CENTER;
			else
				return top ? DividedArea.TOP_RIGHT : DividedArea.BOTTOM_RIGHT;
		}

		//
		public void CanAlwaysTrue(object sender, CanExecuteRoutedEventArgs e) {
			e.CanExecute = true;
		}

        // 次
		public void NextCommand(object sender, ExecutedRoutedEventArgs e) {
			DrawPicture(PageTurnRequest.NEXT);
		}

        // 前
		public void PrevCommand(object sender, ExecutedRoutedEventArgs e) {
			DrawPicture(PageTurnRequest.PREV);
		}

        // ひとつ次
		public void NextOneCommand(object sender, ExecutedRoutedEventArgs e) {
			DrawPicture(PageTurnRequest.NEXT_ONE);
		}

        // ひとつ前
		public void PrevOneCommand(object sender, ExecutedRoutedEventArgs e) {
			DrawPicture(PageTurnRequest.PREV_ONE);
		}

        public void FullCommand(object sender, ExecutedRoutedEventArgs e) {
            Full(!Prefs.data.full);
        }

        // クライアント矩形原点の論理座標をスクリーン原点の論理座標に変換
        Point PointToLogicalScreen(Point p) {
            var q = PointToScreen(p);
            return new Point(q.X / magnifX, q.Y / magnifY);
        }

        // ダイアログ（Window）の位置をLeftとTopに値を指定することで決定する。
        // はたして、このLeftとTopに要求される値は何か。
        // 画面の実際のピクセル数の座標を物理座標、倍率換算された座標を論理座標とする。
        // 物理座標800は125パーセント（120dpi）設定での論理座標640。
        // どうやらWindowのLeftとTopに要求されるのは、スクリーン原点の論理座標らしい。
        // マウスでクリックされた座標にダイアログ（Window）を出したいのだが、
        // どうやらマウスをクリックしたときに得られる座標は、クライアント矩形原点の論理座標のようだ。
        // MSDNでは「指定した要素に対する相対的なマウス ポインターの位置を返します。」となっているので、合っている。
        // 問題はPointToScreenメソッドで、これがスクリーン原点の論理座標に変換してくれるなら良いのだが、
        // どうもスクリーン原点の物理座標を返すようだ。
        // 必要なのはクライアント矩形原点の論理座標をスクリーン原点の論理座標に変換する方法。
        // スクリーンがWindowのような何かとして得られてPointFromScreenが呼べれば解決なのにそういうものはなさげ。
        // 結局まともな方法では得られない倍率を使って計算するしかないってことなんだろうか。解せない。
        // なので、こんなくだらないことに面倒な計算が必要になる。何かが間違ってる。
        public void JumpCommand(Point mousePoint) {
			if (jumpDlg == null) {
				if (!IsEnablePool()) {
					return;
				}
				jumpDlg = new JumpDlg();
				jumpDlg.Owner = this;
				jumpDlg.mainView = this;
				jumpDlg.Setup();
                // X軸方向の計算
                // クライアント矩形のX軸方向真ん中にポップアップさせる。
                // WPFではWindow自身からいわゆるクライアント矩形を得ることはできない。
                // マウスの座標はクライアント矩形原点で来るのにである。
                // グリッドコントロールをクライアント矩形と同じになるように配置してるのは開発者の都合。
                // やはり何かが間違っている。
                // WindowのActualHeightとActualWidthで得られるのは、タイトルバーや枠を含めたサイズ。
                // それが必要な場面もあるのかもしれないが、クライアント矩形ほど重要ではない。
                var x = (gridMain.ActualWidth - jumpDlg.Width) / 2;
                if (x < 0) {
                    x = 0;
                }
                // Y軸方向の計算。
                // Y軸はクリックされた場所にポップアップさせる。
                double y = mousePoint.Y;
                if (mousePoint.Y < 0) {
                    // マウスクリックでの呼び出しではない。
                    y = gridMain.ActualHeight / 2;
                }
                if (WindowState == WindowState.Maximized) {
                    // 全画面表示ではみ出る場合の矯正。
                    if (y + jumpDlg.Height > gridMain.ActualHeight) {
                        y = gridMain.ActualHeight - jumpDlg.Height;
                    }
                }
                // クライアント矩形原点の論理座標をスクリーン原点の論理座標に変換
                var xy = PointToLogicalScreen(new Point(x, y));
                jumpDlg.Left = xy.X;
                jumpDlg.Top = xy.Y;
				// Windows 10にてモーダルダイアログの動きが不可解なので、モードレスにする。
				// おかげで色々と面倒なことに対処する必要が生じる。
				jumpDlg.Show();
				//jumpDlg.ShowDialog();
				//jumpDlg = null;
			}
			else {
                jumpDlg.Focus();
			}
		}

		void FilesCommand() {
            if (filesDlg == null) {
                filesDlg = new FilesDlg();
                filesDlg.Owner = this;
                filesDlg.mainView = this;
                filesDlg.Show();
            }
            else {
                filesDlg.Act();
            }
        }

		// 開く
		// WPFにはディレクトリを指定するダイアログがない。
		// ディレクトリはD&DやSendToを使えばいいやという程度
		void OpenCommand() {
			OpenFileDialog dlg = new OpenFileDialog();
            if (Directory.Exists(Prefs.data.fileOpenDir)) {
                dlg.InitialDirectory = Prefs.data.fileOpenDir;
            }
            dlg.FileName = "";
            dlg.Filter = Properties.Resources.r_FileFilterName + "|*.jpg;*jpeg;*.png;*.gif;*.bmp;*.tif;*.tiff;*.zip;*.cbz";
			if (dlg.ShowDialog(this) == true) {
				Prefs.data.fileOpenDir = Path.GetDirectoryName(dlg.FileName);
				OpenFileFresh(dlg.FileName);
			}
		}

        // 単体表示
        void SingleCommand() {
			Prefs.data.single = !Prefs.data.single;
			DrawPicture(PageTurnRequest.CURRENT);
		}

        // 単純にContextMenu.FindNameを呼ぶだけではアイテムが得られない。
        // どうやったら得られるかは不明。
        // WPFのコンテキストメニューは不可解なことだらけ。
        MenuItem ContextMenuFindName(String name) {
			foreach (MenuItem item in ContextMenu.Items) {
				if (item.Name == name)
					return item;
			}
			return null;
		}

        // 履歴をクリック
		private void OpenRecent_Clicked(object sender, RoutedEventArgs e) {
			var h = Prefs.data.history;
			for (var i = 0; i < h.Count; i++) {
				var ss = h[i].Split(new Char[] { '|' });
				var path = "";
				var page = 0;
				if (ss.Length > 1) {
					path = ss[0];
					try {
						page = Int32.Parse(ss[1]);
					}
					catch (Exception) {
						page = 0;
					}
				}
				if (String.Compare(((String)((MenuItem)sender).Header), path, true) == 0) {
					OpenFile(path, page);
					break;
				}
			}
		}

        // コンテキストメニューの調整
		private void OpenContextMen() {
            // 履歴の調整
            MenuItem opr = ContextMenuFindName("menuItemOpenRecent");
			opr.Items.Clear();
            var h = Prefs.data.history;
            for (var i = 0; i < h.Count; i++) {
                var m = new MenuItem();
                m.Header = h[i].Split(new Char[] { '|' })[0];
                m.Click += OpenRecent_Clicked;
                opr.Items.Add(m);
            }
            // その他調整
            foreach (Object obj in ContextMenu.Items) {
                if (obj.GetType() == typeof(MenuItem)) {
                    var item = (MenuItem)obj;
                    switch (item.Name) {
                        case "menuItemFull":
                            item.IsChecked = Prefs.data.full;
                            break;
                        case "menuItemSingle":
                            item.IsChecked = Prefs.data.single;
                            break;
                        case "menuItemLtoR":
                            item.IsChecked = Prefs.data.ltor;
                            break;
                        case "menuItemAlphabetical":
                            item.IsChecked = Prefs.data.alphabetical;
                            break;
						case "menuItemPageNumber":
							item.IsChecked = Prefs.data.pageNumber;
							break;
                        case "menuItemWithSubdirectory":
                            item.IsChecked = Prefs.data.withSubdirectory;
                            break;
                        case "menuItemGrayScale":
                            item.IsChecked = Prefs.data.grayScale;
                            break;
						case "menuItemHistory":
							item.IsChecked = Prefs.data.enableHistory;
							break;
						case "menuItemGo":
                        case "menuItemNext":
                        case "menuItemPrev":
                        case "menuItemNextOne":
                        case "menuItemPrevOne":
                            item.IsEnabled = IsEnablePool();
                            break;
                    }
                }
            }
        }

        // メニューを開く
        void ContextMenuCommand() {
			OpenContextMen();
			ContextMenu.IsOpen = true;
		}

        // 左から右
		void LtoRCommand() {
			Prefs.data.ltor = !Prefs.data.ltor;
			DrawPicture(PageTurnRequest.CURRENT);
		}

        // アルファベット順
		void AlphabeticalCommand() {
			Prefs.data.alphabetical = !Prefs.data.alphabetical;
            if (IsEnablePool()) {
                ProgressOnTitle();
                pool.ChangeSort(Prefs.data.alphabetical);
            }
            DrawPicture(PageTurnRequest.CURRENT);
        }

        // ページ番号表示
        void PageNumberCommand() {
			Prefs.data.pageNumber = !Prefs.data.pageNumber;
			DrawPicture(PageTurnRequest.CURRENT);
		}

        // サブディレクトリを含む
        void WithSubdirectoryCommand() {
            Prefs.data.withSubdirectory = !Prefs.data.withSubdirectory;
            if (IsEnablePool()) {
                ProgressOnTitle();
                if (jumpDlg != null) {
                    jumpDlg.WaitActor();
                }
                pool.ChangeWithSubdirectory(Prefs.data.withSubdirectory, Prefs.data.alphabetical);
                if (HasContentPool()) {
                    if (Prefs.data.page > pool.GetCount() - 1) {
                        Prefs.data.page = pool.GetCount() - 1;
                    }
                }
                else {
                    Prefs.data.page = 0;
                }
                if (jumpDlg != null) {
                    jumpDlg.Setup();
                }
            }
            DrawPicture(PageTurnRequest.CURRENT);
        }

        void GrayScaleCommand() {
            Prefs.data.grayScale = !Prefs.data.grayScale;
            DrawPicture(PageTurnRequest.CURRENT);
        }

        // 全画面表示
        void Full(bool bFull) {
            if (bFull) {
                //Topmost = true;
                Prefs.data.full = true;
                Prefs.data.windowState = WindowState;
                ResizeMode = ResizeMode.NoResize;
                //WindowState = WindowState.Normal;
                WindowStyle = WindowStyle.None;
                WindowState = WindowState.Maximized;
            }
            else {
                //Topmost = false;
                Prefs.data.full = false;
                ResizeMode = ResizeMode.CanResize;
                WindowStyle = WindowStyle.SingleBorderWindow;
                WindowState = Prefs.data.windowState;
            }
        }

        // 最小化されていたときは元に戻す
        public void Act() {
            if (WindowState == WindowState.Minimized) {
                WindowState = Prefs.data.windowState;
            }
            Activate();
        }

        //-----------------------------------------------------
        // 以下はコントロールからの直接のイベントのコールバック
        //-----------------------------------------------------

        private void mainView_KeyDown(object sender, KeyEventArgs e) {
			// CommandBindingsでは以下のような単独キーの登録はできない。
			if (e.Key == Key.N) {
				NextCommand(sender, null);
			}
			else if (e.Key == Key.B) {
				PrevCommand(sender, null);
			}
			else if (e.Key == Key.Q) {
				Close();
			}
			else if (e.Key == Key.F) {
				Full(!Prefs.data.full);
			}
			else if (e.Key == Key.A) {
				WindowState = WindowState.Minimized;
			}
			else if (e.Key == Key.G) {
				JumpCommand(new Point(-1.0, -1.0));
			}
			else if (e.Key == Key.O) {
				OpenCommand();
			}
			else if (e.Key == Key.S) {
				SingleCommand();
			}
			else if (e.Key == Key.M) {
				ContextMenuCommand();
			}
			else if (e.Key == Key.L) {
				FilesCommand();
			}
			e.Handled = true;
		}

		private void mainView_MouseLeftButtonDown(object sender, MouseButtonEventArgs e) {
			Trace.WriteLine("mainView_MouseLeftButtonDown");
            Activate();
			var w = ActualWidth;
			var h = ActualHeight;
			var p = e.GetPosition(this);
			//Trace.WriteLine(String.Format("{0} {1}, {2}", w, h, p));
			var mpn = GetMousePosDividedArea(w, h, p);
			//Trace.WriteLine(mpn);
			if (mpn == DividedArea.TOP_CENTER) {
				ContextMenuCommand();
			}
			else if (mpn == DividedArea.BOTTOM_CENTER) {
				JumpCommand(p);
			}
			else if (mpn == DividedArea.BOTTOM_LEFT) {
				if (Prefs.data.ltor) {
					PrevCommand(sender, null);
				}
				else {
					NextCommand(sender, null);
				}
			}
			else if (mpn == DividedArea.BOTTOM_RIGHT) {
				if (Prefs.data.ltor) {
					NextCommand(sender, null);
				}
				else {
					PrevCommand(sender, null);
				}
			}
            //Topmost = false;
        }

        private void mainView_MouseDoubleClick(object sender, MouseButtonEventArgs e) {
			Trace.WriteLine("mainView_MouseDoubleClick");
            Activate();
			var mpn = GetMousePosDividedArea(RenderSize.Width, RenderSize.Height, e.GetPosition(this));
			if (mpn == DividedArea.TOP_LEFT) {
				menuItemFull_Click(sender, null);
			}
			else if (mpn == DividedArea.TOP_RIGHT) {
            }
		}

		private void mainView_MouseRightButtonDown(object sender, MouseButtonEventArgs e) {
			Trace.WriteLine("mainView_MouseRightButtonDown");
		}

		private void menuItemOpen_Click(object sender, RoutedEventArgs e) {
			OpenCommand();
		}

		private void menuItemFiles_Click(object sender, RoutedEventArgs e) {
			FilesCommand();
		}

		private void mainView_Closing(object sender, CancelEventArgs e) {
			if (pool != null) {
                pool.Join();
			}
            if (jumpDlg != null) {
                jumpDlg.Close();
            }
            if (filesDlg != null) {
                filesDlg.Close();
            }
            FixHistory();
			if (WindowState == WindowState.Normal) {
				Prefs.data.windowTop = Top;
				Prefs.data.windowLeft = Left;
				Prefs.data.windowWidth = Width;
				Prefs.data.windowHeight = Height;
			}
			if (!Prefs.data.full) {
				Prefs.data.windowState = WindowState;
			}
			Prefs.data.Save();
		}

		private void menuItemNext_Click(object sender, RoutedEventArgs e) {
			NextCommand(sender, null);
		}

		private void menuItemPrev_Click(object sender, RoutedEventArgs e) {
			PrevCommand(sender, null);
		}

		private void menuItemExit_Click(object sender, RoutedEventArgs e) {
			Close();
		}

		private void menuItemFull_Click(object sender, RoutedEventArgs e) {
			Full(!Prefs.data.full);
		}

		private void mainView_SizeChanged(object sender, SizeChangedEventArgs e) {
			//Trace.WriteLine("mainView_SizeChanged");
			DrawPicture(PageTurnRequest.CURRENT);
		}

		private void menuItemNextOne_Click(object sender, RoutedEventArgs e) {
			DrawPicture(PageTurnRequest.NEXT_ONE);
		}

		private void menuItemPrevOne_Click(object sender, RoutedEventArgs e) {
			DrawPicture(PageTurnRequest.PREV_ONE);
		}

		private void menuItemGo_Click(object sender, RoutedEventArgs e) {
			JumpCommand(new Point(-1.0, -1.0));
		}

		private void menuItemMinimize_Click(object sender, RoutedEventArgs e) {
			WindowState = WindowState.Minimized;
		}

		private void menuItemSingle_Click(object sender, RoutedEventArgs e) {
			SingleCommand();
		}

		private void mainView_ContextMenuOpening(object sender, ContextMenuEventArgs e) {
			OpenContextMen();
		}

		private void menuItemLtoR_Click(object sender, RoutedEventArgs e) {
			LtoRCommand();
		}

		private void menuItemAlphabetical_Click(object sender, RoutedEventArgs e) {
			AlphabeticalCommand();
		}

		private void menuItemPageNumber_Click(object sender, RoutedEventArgs e) {
			PageNumberCommand();
		}

		private void mainView_Drop(object sender, DragEventArgs e) {
			var args = (string[])e.Data.GetData(DataFormats.FileDrop, false);
			if (args.Length > 0) {
				OpenFileFresh(args[0]);
			}
		}

        private void menuItemWithSubdirectory_Click(object sender, RoutedEventArgs e) {
            WithSubdirectoryCommand();
        }

        private void menuItemGrayScale_Click(object sender, RoutedEventArgs e) {
            GrayScaleCommand();
        }

		private void menuItemHistory_Click(object sender, RoutedEventArgs e) {
			Prefs.data.enableHistory = !Prefs.data.enableHistory;
		}

		private void menuItemAbout_Click(object sender, RoutedEventArgs e) {
			if (aboutDlg == null) { 
				aboutDlg = new AboutDlg();
				aboutDlg.mainView = this;
				aboutDlg.Owner = this;
				aboutDlg.Show();
			}
			else {
				aboutDlg.Focus();
			}
		}

        Boolean activatedForStateChange = false;

        private void mainView_Activated(object sender, EventArgs e) {
            //Trace.WriteLine("mainView_Activated");
            activatedForStateChange = true;
            InputMethod.SetIsInputMethodEnabled(this, false);
        }

        private void mainView_Deactivated(object sender, EventArgs e) {
            //Trace.WriteLine("mainView_Deactivated");
            activatedForStateChange = false;
        }

        private void mainView_StateChanged(object sender, EventArgs e) {
            //var name = WindowState.GetName(typeof(System.Windows.WindowState), WindowState);
            //Trace.WriteLine("mainView_StateChanged: " + name);
            if (activatedForStateChange) {
                activatedForStateChange = false;
                if (WindowState == WindowState.Maximized && Prefs.data.full) {
                    ResizeMode = ResizeMode.NoResize;
                    //WindowState = WindowState.Normal;
                    WindowStyle = WindowStyle.None;
                    WindowState = WindowState.Maximized;
                }
            }
        }
    }

}
