﻿/*
	PageBitmap.cs
	Copyright 2016, OSTRA / Takeshi Yoneki
*/
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace PictureSetViewLite {
	// 指定サイズに収まるよう拡大縮小した画像のストリームを得る。
	public class PageBitmap {
		// 動作環境
		// GetPageBitmapStream呼出し後、pageとpageTypeは値が変更される。
		public int page = 0;
		public PageType pageType = PageType.NONE;

		public PageBitmap(int page, PageType pageType) {
			this.page = page;
			this.pageType = pageType;
		}

        bool isPortrait(int width, int height) {
            return height > width;
        }

		// ページ進行に合わせ画像ストリームを得る。
		// PageTurnRequest.CURRENT指定時はなるべく値を返す。
		// それ以外は既に最後のページであるとかの理由でnullを返す。
		// C#がスマートに複数戻り値を返すことが出来るようになるまでは、オブジェクトインスタンスの変数に頼る。
		public MemoryStream GetPageBitmapStream(BitmapPool pool, PageTurnRequest turn, int width, int height, bool ltor, bool single) {
			if (pool == null || !pool.IsFileExist()) {
				return null;
			}
			if (single) {
				if (turn == PageTurnRequest.CURRENT) {
				}
				else if (turn == PageTurnRequest.NEXT || turn == PageTurnRequest.NEXT_ONE) {
					if (page + 1 < pool.GetCount()) {
						page += 1;
					}
					else {
						return null;
					}
				}
				else if (turn == PageTurnRequest.PREV || turn == PageTurnRequest.PREV_ONE) {
					if (page - 1 >= 0) {
						page -= 1;
					}
					else {
						return null;
					}
				}
				var bm = pool.GetNth(page);
				pageType = PageType.ONE;
				return PageBitmap.GetOneBitmapStream(bm, width, height);
			}
			// 縦画面
			if (isPortrait(width, height)) {
				if (turn == PageTurnRequest.NEXT) {
					if (pageType == PageType.ONE ||
						pageType == PageType.TWO ||
						pageType == PageType.END) {
						if (page + 1 < pool.GetCount()) {
							page += 1;
						}
						else {
							return null;
						}
					}
				}
				else if (turn == PageTurnRequest.NEXT_ONE) {
					if (page + 1 < pool.GetCount()) {
						page += 1;
					}
					else {
						return null;
					}
				}
				else if (turn == PageTurnRequest.PREV) {
					if (pageType == PageType.ONE ||
						pageType == PageType.TWO ||
						pageType == PageType.START) {
						if (page - 1 >= 0) {
							page -= 1;
						}
						else {
							return null;
						}
					}
				}
				else if (turn == PageTurnRequest.PREV_ONE) {
					if (page - 1 >= 0) {
						page -= 1;
					}
					else {
						return null;
					}
				}
				var bm_src = pool.GetNth(page);
				if (bm_src == null) {
                    pageType = PageType.ONE;
                    return null;
				}
				// 縦画像
				else if (isPortrait(bm_src.Size.Width, bm_src.Size.Height)) {
					pageType = PageType.ONE;
					return PageBitmap.GetOneBitmapStream(bm_src, width, height);
				}
				// 横画像
				else {
					if (turn == PageTurnRequest.CURRENT) {
						if (pageType == PageType.END) {
							pageType = PageType.END;
							return PageBitmap.GetEndBitmapStream(bm_src, width, height, ltor);
						}
						else {
							pageType = PageType.START;
							return PageBitmap.GetStartBitmapStream(bm_src, width, height, ltor);
						}
					}
					else if (turn == PageTurnRequest.NEXT || turn == PageTurnRequest.NEXT_ONE) {
						if (pageType != PageType.START || turn == PageTurnRequest.NEXT_ONE) {
							pageType = PageType.START;
							return PageBitmap.GetStartBitmapStream(bm_src, width, height, ltor);
						}
						else {
							pageType = PageType.END;
							return PageBitmap.GetEndBitmapStream(bm_src, width, height, ltor);
						}
					}
					else { // PREV, PREV_ONE
						if (pageType == PageType.END || turn == PageTurnRequest.PREV_ONE) {
							pageType = PageType.START;
							return PageBitmap.GetStartBitmapStream(bm_src, width, height, ltor);
						}
						else {
							pageType = PageType.END;
							return PageBitmap.GetEndBitmapStream(bm_src, width, height, ltor);
						}
					}
				}
			}
			// 横画面
			else {
				var bReqOne = false;
				if (turn == PageTurnRequest.NEXT) {
					if (pageType == PageType.TWO) {
						if (page + 2 < pool.GetCount()) {
							page += 2;
						}
						else {
							return null;
						}
					}
					else {
						if (page + 1 < pool.GetCount()) {
							page += 1;
						}
						else {
							return null;
						}
					}
				}
				else if (turn == PageTurnRequest.NEXT_ONE) {
					if (page + 1 < pool.GetCount()) {
						page += 1;
					}
					else {
						return null;
					}
				}
				else if (turn == PageTurnRequest.PREV) {
					if (page - 2 >= 0) {
						var bm = pool.GetNth(page - 1);
						if (bm == null) {
							return null;
						}
						else if (!isPortrait(bm.Width, bm.Height)) {
							page -= 1;
							bReqOne = true;
						}
						else {
							bm = pool.GetNth(page - 2);
							if (bm == null) {
								return null;
							}
							else if (!isPortrait(bm.Width, bm.Height)) {
								page -= 1;
								bReqOne = true;
							}
							else {
								page -= 2;
							}
						}
					}
					else if (page - 1 >= 0) {
						page -= 1;
						bReqOne = true;
					}
					else {
						return null;
					}
				}
				else if (turn == PageTurnRequest.PREV_ONE) {
					if (page - 1 >= 0) {
						page -= 1;
					}
					else if (pageType == PageType.TWO) {
                        page = 0;
                        bReqOne = true;
                    }
                    else {
                        return null;
                    }
                }
				var bm_start = pool.GetNth(page);
				if (bm_start == null) {
                    pageType = PageType.ONE;
                    return null;
				}
				else if (bReqOne) {
					pageType = PageType.ONE;
					return PageBitmap.GetOneBitmapStream(bm_start, width, height);
				}
				// 縦画像
				else if (isPortrait(bm_start.Size.Width, bm_start.Size.Height)) {
					var bm_end = pool.GetNth(page + 1);
					if (bm_end != null && isPortrait(bm_end.Size.Width, bm_end.Size.Height)) {
						pageType = PageType.TWO;
						return PageBitmap.GetTwoBitmapStream(bm_start, bm_end, width, height, ltor);
					}
					else {
						pageType = PageType.ONE;
						return PageBitmap.GetOneBitmapStream(bm_start, width, height);
					}
				}
				// 横画像
				else {
					pageType = PageType.ONE;
					return PageBitmap.GetOneBitmapStream(bm_start, width, height);
				}
			}
		}

		// 指定サイズに収まるよう拡大縮小した画像のストリームを得る
		static MemoryStream GetMagnifiedBitmapStream(Bitmap bm_src, int width, int height) {
			if (bm_src == null) {
				return null;
			}
			var mp = new MemoryStream();
			var w = bm_src.Size.Width;
			var h = bm_src.Size.Height;
			double mg = 1.0;
			if (width == 0 || (w == width && h < height) || (w < width && h == height)) {
				// 一致してるときのみそのまま。
				bm_src.Save(mp, System.Drawing.Imaging.ImageFormat.Bmp);
			}
			else {
				mg = Math.Min((double)height / (double)h, (double)width / (double)w);
				h = (int)(mg * h);
				w = (int)(mg * w);
				var bm_dst = new Bitmap(w, h);
				var g = Graphics.FromImage(bm_dst);
				g.Clear(Color.LightGray);
				g.DrawImage(bm_src, 0, 0, w, h);
				bm_dst.Save(mp, System.Drawing.Imaging.ImageFormat.Bmp);
				bm_dst.Dispose();
			}
			mp.Seek(0, SeekOrigin.Begin);
			return mp;
		}

		// 右半分
		static MemoryStream GetRightBitmapStream(Bitmap bm_src, int width, int height) {
			if (bm_src == null) {
				return null;
			}
			var w = bm_src.Size.Width / 2;
			var h = bm_src.Size.Height;
			var r = new System.Drawing.Rectangle(new System.Drawing.Point(w, 0), new System.Drawing.Size(w, h));
			var bm_cut = bm_src.Clone(r, System.Drawing.Imaging.PixelFormat.DontCare);
			var mp = GetMagnifiedBitmapStream(bm_cut, width, height);
			bm_cut.Dispose();
			return mp;
		}

		// 左半分
		static MemoryStream GetLeftBitmapStream(Bitmap bm_src, int width, int height) {
			if (bm_src == null) {
				return null;
			}
			var w = bm_src.Size.Width / 2;
			var h = bm_src.Size.Height;
			var r = new System.Drawing.Rectangle(new System.Drawing.Point(0, 0), new System.Drawing.Size(w, h));
			var bm_cut = bm_src.Clone(r, System.Drawing.Imaging.PixelFormat.DontCare);
			var mp = GetMagnifiedBitmapStream(bm_cut, width, height);
			bm_cut.Dispose();
			return mp;
		}

		// 前半
		static MemoryStream GetStartBitmapStream(Bitmap bm_src, int width, int height, bool ltor) {
			if (ltor) {
				return GetLeftBitmapStream(bm_src, width, height);
			}
			else {
				return GetRightBitmapStream(bm_src, width, height);
			}
		}

		// 後半
		static MemoryStream GetEndBitmapStream(Bitmap bm_src, int width, int height, bool ltor) {
			if (ltor) {
				return GetRightBitmapStream(bm_src, width, height);
			}
			else {
				return GetLeftBitmapStream(bm_src, width, height);
			}
		}

		// 全体
		static MemoryStream GetOneBitmapStream(Bitmap bm_src, int width, int height) {
			if (bm_src == null) {
				return null;
			}
			return GetMagnifiedBitmapStream(bm_src, width, height);
		}

		// 二つ並べて
		static MemoryStream GetLeftRightBitmapStream(Bitmap bm_left, Bitmap bm_right, int width, int height) {
			if (bm_right == null || bm_left == null) {
				return null;
			}
			var h = Math.Max(bm_right.Size.Height, bm_left.Size.Height);
			double mg_left = (double)h / (double)bm_left.Size.Height;
			double mg_right = (double)h / (double)bm_right.Size.Height;
			var w = bm_right.Size.Width * mg_right + bm_left.Size.Width * mg_left;
			double mg = 1.0;
			if (width != 0) {
				mg = Math.Min((double)height / (double)h, (double)width / (double)w);
				h = (int)(mg * h);
				w = (int)(mg * w);
			}
			var bm_dst = new Bitmap((int)w, (int)h);
			var g = Graphics.FromImage(bm_dst);
			g.Clear(Color.LightGray);
			var h_left = (int)(mg * mg_left * bm_left.Height);
			var h_right = (int)(mg * mg_right * bm_right.Height);
			var w_left = (int)(mg * mg_left * bm_left.Width);
			var w_right = (int)(mg * mg_right * bm_right.Width);
			g.DrawImage(bm_left, 0, 0, w_left, h_left);
			g.DrawImage(bm_right, w_left, 0, w_right, h_right);
			var mp = new MemoryStream();
			bm_dst.Save(mp, System.Drawing.Imaging.ImageFormat.Bmp);
			mp.Seek(0, SeekOrigin.Begin);
			bm_dst.Dispose();
			return mp;
		}

		// 二つ並べて
		static MemoryStream GetTwoBitmapStream(Bitmap bm_start, Bitmap bm_end, int width, int height, bool ltor) {
			if (ltor) {
				return PageBitmap.GetLeftRightBitmapStream(bm_start, bm_end, width, height);
			}
			else {
				return PageBitmap.GetLeftRightBitmapStream(bm_end, bm_start, width, height);
			}
		}
	}
}
