﻿using System;
using System.Xml.Serialization;
using System.IO;
using System.Windows;
using System.Windows.Media;
using System.Collections.Specialized;

namespace PictureSetViewLite {
    public static class Prefs {
        public static Preferences data = Preferences.Load();
    }

    public class Preferences {
        public string fileOpenDir = "";
        public string filePath = "";
        public double windowTop = 0;
        public double windowLeft = 0;
        public double windowWidth = 0;
        public double windowHeight = 0;
        public WindowState windowState = WindowState.Normal;
        public int page = 0;
        public Stretch stretch = Stretch.Uniform;
        public bool full = false;
        public bool single = false;
        public StringCollection history;
        public bool ltor = false;
        public bool alphabetical = false;
        public bool pageNumber = true;
        public bool withSubdirectory = false;
        public bool grayScale = false;
        public int filesScriptIndex = 0;
        public double filesDlgTop = 0;
        public double filesDlgLeft = 0;
        public double filesDlgWidth = 0;
        public double filesDlgHeight = 0;
        public int fileItemIndex = 0;
        public bool filesDlgClose = true;
        public WindowState filesDlgWindowState = WindowState.Normal;
        public bool enableHistory = true;

        static string GetSettingDirPath() {
            return Path.Combine(System.Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "OSTRA");
        }

        static string GetSettingFilePath() {
            return Path.Combine(GetSettingDirPath(), "PictureSetViewLite_1_0.pref");
        }

        public static Preferences Load() {
            XmlSerializer serializer = new XmlSerializer(typeof(Preferences));
            FileStream fs = null;
            try {
                fs = new FileStream(GetSettingFilePath(), FileMode.Open);
            }
            catch {
                fs = null;
            }
            Preferences result = null;
            if (fs != null) {
                try {
                    result = (Preferences)serializer.Deserialize(fs);
                }
                catch {
                }
                fs.Close();
            }
            if (result == null) {
                result = new Preferences();
            }
            return result;
        }

        public bool Save() {
            bool result = false;
            string dir = GetSettingDirPath();
            if (Directory.Exists(dir) == false) {
                Directory.CreateDirectory(dir);
            }
            XmlSerializer serializer = new XmlSerializer(typeof(Preferences));
            FileStream fs = new FileStream(GetSettingFilePath(), FileMode.Create);
            if (fs != null) {
                serializer.Serialize(fs, this);
                fs.Close();
                result = true;
            }
            return result;
        }
    }
}